package org.eso.ohs.core.dbb.rmi;

// $Id: DbbRemoteEngineImpl.java,v 1.2 2004/02/17 14:31:57 tcanavan Exp $

import java.net.MalformedURLException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.sql.SQLException;

import org.eso.ohs.core.dbb.server.DbbQuery;
import org.eso.ohs.core.dbb.server.DbbSqlEngine;
import org.eso.ohs.core.dbb.sql.DbbSqlChunk;
import org.eso.ohs.core.dbb.sql.DbbSqlQueryChunks;
import org.eso.ohs.core.dbb.sql.DbbSqlTable;
import org.eso.ohs.core.utilities.Logger;


/**   
 * Implementation of the remote sql engine.
 *
 * @author Bob Kemp
 * <br>ChangeLog: C. Boarotto, ESO 10-Sep-2000
 * <br>ChangeLog: amchavan, 17-Oct-2001: added logging
 */
public class DbbRemoteEngineImpl 
    extends UnicastRemoteObject
    implements DbbRemoteEngine {

    /** Sql engine reference */
    private DbbSqlEngine dbbQ_;

    /** Where logging output goes */
    private Logger logger_;

    /** Host name of the client connected to this remote engine. */
    private String clientHost_;


    //----------------------------------------
    // Constructor
    //----------------------------------------

    /**
     * Construct a remote engine implementation 
     *
     * @param dbbQ reference to the sql engine to generate the query
     *
     * @param logger    Where logging output goes. If this is 
     *                  <code>null</code>, no logging is generated.
     *
     * @param clientHost    Host name of the client connected 
     *                      to this remote engine.
     *
     * <br>ChangeLog: amchavan, 17-Oct-2001: added logging
     */
    public DbbRemoteEngineImpl( DbbSqlEngine dbbQ, 
				Logger logger, 
				String clientHost )
	throws RemoteException {
	dbbQ_       = dbbQ;
	logger_     = logger;
	clientHost_ = clientHost;
    }


    /**
     * Construct a remote engine implementation 
     *
     * @param dbbQ reference to the sql engine to generate the query
     */
    public DbbRemoteEngineImpl(DbbSqlEngine dbbQ)
	throws RemoteException {
	this( dbbQ, null, null );
    }


    //----------------------------------------
    // Public methods
    //----------------------------------------

    /**
     * Create and start a thread to get the result set and to process it
     * into an Object array.  
     *
     * @param chunks DbbSqlQueryChunks
     *
     * @exception SQLException JDBC exception occurs
     * @exception RemoteException RMI exception occurs
     * @exception MalformedURLException malformed url i.e. servername//port
     * @exception NotBoundException client could not bind to server
     */
    public DbbQuery startQuery(DbbSqlQueryChunks chunks)
	throws SQLException, RemoteException,
	MalformedURLException, NotBoundException {

	log( "Queried from " + clientHost_ );
	return new DbbRemoteQueryImpl(dbbQ_.startQuery(chunks));
    }


    /**
     * Create and start a thread to exec the sql count action.
     *
     * @param chunks DbbSqlQueryChunks
     *
     * @exception SQLException JDBC exception occurs
     * @exception RemoteException RMI exception occurs
     * @exception MalformedURLException malformed url i.e. servername//port
     * @exception NotBoundException client could not bind to server
     */
    public int execSqlCount(DbbSqlQueryChunks chunks)
	throws SQLException, RemoteException,
	MalformedURLException, NotBoundException {

	return dbbQ_.execSqlCount(chunks);
    }


    /**
     * Add default colums to a particular chunk
     *
     * @param chunk Sql Chunk
     *
     * @exception SQLException JDBC exception occurs
     * @exception RemoteException RMI exception occurs
     * @exception MalformedURLException malformed url i.e. servername//port
     * @exception NotBoundException client could not bind to server
     */
    public void addDefaultColumnsChunk(DbbSqlChunk chunk)
	throws SQLException, RemoteException,
	MalformedURLException, NotBoundException {
	dbbQ_.addDefaultColumnsChunk(chunk);
    }

    public DbbSqlChunk[] getDefaultColumnsChunk()
	throws SQLException, RemoteException,
	MalformedURLException, NotBoundException {
	return dbbQ_.getDefaultColumnsChunk();
    }

    /**
     * Add a preferential join condition to be considered when building
     * automatically the query.
     *
     * @param from_tbl	DbbSqlTable indicating the primary key table.
     * @param from_col	String indicating the primary key table column.
     * @param to_tbl	DbbSqlTable indicating the foreign key table.
     * @param to_col	String indicating the foreign key table column.
     *
     * @exception SQLException JDBC exception occurs
     * @exception RemoteException RMI exception occurs
     * @exception MalformedURLException malformed url i.e. servername//port
     * @exception NotBoundException client could not bind to server
     */
    public void addJoinCondition(DbbSqlTable from_tbl,
				 String from_col,
				 DbbSqlTable to_tbl,
				 String to_col)
	throws SQLException, RemoteException,
	MalformedURLException, NotBoundException {

	dbbQ_.addJoinCondition(from_tbl, from_col,
			       to_tbl, to_col);
    }

    /**
     * Add a left outer join condition to be considered when building
     * automatically the query.
     *
     * @param from_tbl  DbbSqlTable indicating the primary key table.
     * @param from_col  String indicating the primary key table column.
     * @param to_tbl    DbbSqlTable indicating the foreign key table.
     * @param to_col    String indicating the foreign key table column.
     *
     * @exception SQLException JDBC exception occurs
     * @exception RemoteException RMI exception occurs
     * @exception MalformedURLException malformed url i.e. servername//port
     * @exception NotBoundException client could not bind to server
     */
    public void addCompositeJoinCondition(DbbSqlTable from_tbl,
                 String from_cols[],
                 DbbSqlTable to_tbl,
                 String to_cols[], boolean leftOuterJoin)
            throws SQLException, RemoteException,
            MalformedURLException, NotBoundException {

        dbbQ_.addCompositeJoinCondition(from_tbl, from_cols,
                       to_tbl, to_cols, leftOuterJoin);
    }


    //----------------------------------------
    // Protected methods
    //----------------------------------------

    /**
     * Log a message.<BR>
     * This is a shortcut to the Logger API; if no Logger is defined,
     * this method does nothing.
     *
     * @param msg The message to log.
     */
    public void log( String msg ) {
	if( logger_ != null )
	    logger_.log( msg );
    }


    /**
     * Log an exception.<BR>
     * This is a shortcut to the Logger API; if no Logger is defined,
     * this method does nothing.
     *
     * @param e    The exception to log.
     * @param msg  An extra explanatory message.
     */
    public void log( String msg, Exception e ) {
	if( logger_ != null )
	    logger_.logException( msg, e );
    }
}



