package org.eso.ohs.core.dbb.server;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.ohs.core.dbb.sql.DbbSqlChunk;
import org.eso.ohs.core.dbb.sql.DbbSqlOperator;
import org.eso.ohs.core.dbb.sql.DbbSqlQueryChunks;
import org.eso.ohs.core.dbb.sql.DbbSqlTable;
import org.eso.ohs.core.utilities.MsgManager;

/**
 * This is the main engine of a database browser: builds SQL
 * statements, performs queries (within threads), returns result data.
 * <br>
 * $Id: DbbSqlEngineImpl.java,v 1.4 2005/06/16 15:54:16 gdonino Exp $
 * <br>
 * @version $Revision: 1.4 $
 * @author Carlo Boarotto, ESO 19-Jul-1999
 * <br>
 * ChangeLog: A M Chavan, ESO 26-Jul-1999
 */

public class DbbSqlEngineImpl implements DbbSqlEngine {

    protected static Logger stdlog_ = Logger.getLogger(DbbSqlEngineImpl.class);

    /** Aggregate object to build the join. */
    protected DbbSqlJoinBuilder joiner_ = null;

    /** Link to the database session object */
    private DbbSession session_ = null;

    /** Default set of columns to retrieve. */
    protected DbbSqlChunk[] defColumns_= new DbbSqlChunk[0];

    /** Default set of rows to retrieve. */
    protected DbbSqlChunk[] defRows_;

    /** Default set of sort columns. */
    protected DbbSqlChunk[] defSort_;

    /** Default set of "group by" columns. */
    protected DbbSqlChunk[] defGroup_;

    /** List of tables for the FROM clause */
    protected Vector tables_ = new Vector();

    /** List of WHERE clauses */
    protected Vector whereClauses_ = new Vector();

    //----------------------------------------
    // Constructors
    //----------------------------------------

    /**
     * Build an SQL engine
     *
     * @param db_session Database session
     * @exception SQLException if generating a session happens
     */
    public DbbSqlEngineImpl(DbbSession db_session) throws SQLException {
        session_ = db_session;
        joiner_ = new DbbSqlJoinBuilder(db_session);
    }

    /**
     * Build an SQL engine
     *
     * @param db_session Database session
     *
     * @param get_db_keys Boolean flag: if false use only the key
     *                          information provided by the end application
     *                          to build the sql join clauses
     *
     * @exception SQLException if generating a session fails
     *
     */
    public DbbSqlEngineImpl(DbbSession db_session, boolean get_db_keys)
        throws SQLException {

        session_ = db_session;
        joiner_ = new DbbSqlJoinBuilder(db_session, get_db_keys);
    }

    //----------------------------------------
    // Private Methods
    //----------------------------------------

    /**
     * Build one line of the SELECT clause of an SQL query statement
     *
     * @param chunk  An SQL chunk from which we build the line
     * @param first  If not true, will insert a separating comma
     *
     * @return one line of the SELECT clause
     */
    private String sqlBuildSelectLine(DbbSqlChunk chunk, boolean first) {

        String sql = "";

        /* Add commas between different SQL lines */
        if (!first)
            sql = sql + ",";
        sql = sql + "\n    ";

        /* Use the chunk labels, if any */
        String label = chunk.getLabel();
        if (label != null)
            sql = sql + " " + label + " = ";

        //     String expr = chunk.getExpression();
        //     if( expr != "" && expr != null )
        //       sql = sql + expr;
        //     else

        sql = sql + chunk.getColumnSpec();

        return (sql);
    }

    /**
     * Table list: precompute from the available chunks
     * See sqlBuildQuery for a description of the parameters
     */
    protected void tablesCompute(
            DbbSqlChunk[] columns,
            DbbSqlChunk[] rows,
            DbbSqlChunk[] sort,
            DbbSqlChunk[] group) {

        for (int i = 0; i < columns.length; i++)
            tablesAdd(columns[i].getTable());

        for (int i = 0; i < rows.length; i++)
            tablesAdd(rows[i].getTable());

        for (int i = 0; i < sort.length; i++)
            tablesAdd(sort[i].getTable());

        for (int i = 0; i < group.length; i++)
            tablesAdd(group[i].getTable());
    }

    //----------------------------------------
    // Protected Methods
    //----------------------------------------

    /* @return true if a clause is already defined, false otherwise. */
    protected boolean whereClausesFound(String clause) {
        return whereClauses_.contains(clause);
    }

    /** WHERE clause list: clear all the contents */
    protected void whereClausesClear() {
        whereClauses_.removeAllElements();
    }

    /**
     * WHERE clause list: add a WHERE clause
     *
     * @param clause A WHERE clause to be added to the list
     */
    protected void whereClausesAdd(String clause) {
        if (clause != null && clause.length() > 0 && !whereClauses_.contains(clause))
            whereClauses_.addElement(clause);
    }

    /* @return true if a table is already defined, false otherwise. */
    protected boolean tablesFound(DbbSqlTable table) {
        return tables_.contains(table);
    }

    /** Table list: clear all the contents. */
    protected void tablesClear() {
        tables_.removeAllElements();
    }

    /**
     * Table list: add a table
     *
     * @param table A table object to be added to the list
     */
    protected void tablesAdd(DbbSqlTable table) {

        if (table != null && !tables_.contains(table)) {

            /* Check if the table definition is complete, otherwise take the
             * database name and the owner from the session defaults
             */
            table.completeTable(session_);

            tables_.addElement(table);
        }
    }

    /**
     * Build the SELECT clause of an SQL query statement based on the
     * chunks passed on as arguments to this method
     *
     * @param isCount   Boolean flag to indicate if it is only a sql
     *                          count statement
     *
     * @return The SELECT fragment of an SQL query
     */
    protected String sqlBuildSELECT(DbbSqlChunk[] columns, boolean isCount) {

        String sql = "SELECT DISTINCT ";

        if (isCount)
            return (sql + " count(*)\n");

        /* Test boundary condition */
        if (columns.length == 0)
            return (sql + "*\n");

        for (int i = 0; i < columns.length; i++)
            sql += sqlBuildSelectLine(columns[i], (i == 0));

        return (sql + "\n");
    }

    /** Build the FROM clause of an SQL query statement */
    protected String sqlBuildFROM() {

        int len = tables_.size();
        String sql = "";

        for (int i = 0; i < len; i++) {

            DbbSqlTable table = (DbbSqlTable)tables_.elementAt(i);

            if (!sql.equals(""))
                sql += ",";
            sql += "\n    " + table.get();
        }

        return "FROM" + sql;
    }

    /** Build the WHERE clause of an SQL query statement */
    protected String sqlBuildWHERE() {

        String sql = "";

        // @author amchavan, 10-Jun-2002: commented out
        //if(stdlog_.isDebugEnabled())
        //    dumpWhereClauses( "dumpWhereClauses: sqlBuildWHERE" );

        int len = whereClauses_.size();

        for (int i = 0; i < len; i++) {

            String clause = (String)whereClauses_.elementAt(i);

            // stdlog_.debug("Clause n:"+i+": <"+clause+">");

            if (i == 0)
                sql = "WHERE " + clause + "\n";
            else
                sql = sql + "AND " + clause + "\n";
        }

        return sql;
    }

    /** Build the ORDER BY clause of an SQL query statement */
    protected String sqlBuildSORT(DbbSqlChunk[] sort) {

        String sql = "";

        for (int i = 0; i < sort.length; i++) {
            if (!sql.equals(""))
                sql += ",";
            sql += "\n    " + sort[i].getColumnSpec();

            /* Get the sorting criterion from the sql chunk. */
            DbbSqlOperator op = sort[i].getOperator();
            if (op != null && op.get().equals("descending"))
                sql += " DESC";
        }

        if (!sql.equals(""))
            sql = "ORDER BY" + sql;

        return sql;
    }

    /** Build the GROUP BY clause of an SQL query statement */
    protected String sqlBuildGROUP(DbbSqlChunk[] group) {

        String sql = "";
        // Still TBD
        return sql;
    }

    /**
     * Process the list of chunks in the rows selection panel which
     * were activated by the user, and extract from there all the
     * needed WHERE clauses. Clauses are stored in the WHERE clause
     * list using whereClausesAdd()
     */
    protected void computeRowSelectionClauses(DbbSqlChunk[] rows) {

        String sql = null;
        DbbSqlChunk chunk;

        for (int i = 0; i < rows.length; i++) {

            chunk = rows[i];

            sql = chunk.getSelectClause();
            // sql = chunk.getColumnSpec() + " " + chunk.getOperator().get() +
            //     " " + chunk.getValue();

            // @author amchavan, 10-Jun-2002: commented out
            // stdlog_.debug("computeRowSelectionClauses - SQL: " + sql);

            whereClausesAdd(sql);
        }
    }

    /**
     * Build join conditions navigating through the database tables,
     * following the foreign keys. It is not necessary that all the
     * foreign keys have been defined on the database, since the end
     * application can define its own foreign keys invoking the
     * addDbbSqlKey method. In any case this set of keys is preferred
     * when building a join path between two given tables.
     *
     * @see DbbSqlJoinBuilder
     * @see DbbSqlJoinInfo
     * @exception SQLException if a jgbc call throws an excpetion
     */
    protected void computeTableJoinClauses() throws SQLException {

        DbbSqlJoinInfo join_info = null;
        DbbSqlJoinInfo prev_join = null;

        int len = tables_.size();

        /* Join each couple of tables: is it the right way to do it? */
        for (int i = 0; i < len; i++) {

            DbbSqlTable from_table = (DbbSqlTable)tables_.elementAt(i);
		
            for (int k = i; k < len; k++) {

                DbbSqlTable to_table = (DbbSqlTable)tables_.elementAt(k);

				stdlog_.debug("from table: " + from_table + " to table: " + to_table);
                if (from_table.equals(to_table)) {
					stdlog_.debug("SKIP!");
                    continue;
                }

                if (prev_join != null && prev_join.isContainedTable(to_table)) {
					stdlog_.debug("prevJoin: " + prev_join + "contains table: " + to_table);
                    continue;
                }

                /* @author amchavan, 10-Jun-2002: commented out */
                 stdlog_.debug("Computing join between <" +
                     from_table + "> and <" + to_table + ">");
                

                join_info = joiner_.getJoin(from_table, to_table);

                if (prev_join != null) {
                    prev_join.mergeTables(join_info);
                    prev_join.mergeClauses(join_info);
                } else {
                    prev_join = join_info;
                }
            }
        }

        /* Null if there are no tables to join */
        if (prev_join != null) {
            DbbSqlTable[] join_tables = prev_join.getTables();

            for (int i = 0; i < join_tables.length; i++)
                tablesAdd(join_tables[i]);

            String[] join_clauses = prev_join.getClauses();
            for (int i = 0; i < join_clauses.length; i++)
                whereClausesAdd(join_clauses[i]);
        }
    }

    //----------------------------------------
    // Public methods
    //----------------------------------------

    /**
     * Add default colums to a particular chunk
     *
     * @param chunk Sql Chunk
     *
     */
    public void addDefaultColumnsChunk(DbbSqlChunk chunk) {
        defColumns_ = append(defColumns_, chunk);
    }

    public DbbSqlChunk[] getDefaultColumnsChunk() {
        return defColumns_;
    }

    /**
     * Add a preferential join condition to be considered when building
     * automatically the query
     *
     * @param from_tbl  DbbSqlTable indicating the primary key table
     * @param from_col  String indicating the primary key table column
     * @param to_tbl    DbbSqlTable indicating the foreign key table
     * @param to_col    String indicating the foreign key table column
     */
    public void addJoinCondition(
            DbbSqlTable from_tbl,
            String from_col,
            DbbSqlTable to_tbl,
            String to_col) {

        if (from_col == null || from_col.trim().equals(""))
            return;
        if (to_col == null || to_col.trim().equals(""))
            return;
        if (from_tbl == null || to_tbl == null)
            return;

        from_tbl.completeTable(session_);
        to_tbl.completeTable(session_);

        /* Create a normal key. */
        DbbSqlKey new_key = new DbbSqlKey(null, from_tbl, from_col, to_tbl, to_col);
        joiner_.addDbbSqlKey(new_key);
    }

    /**
     * Add a preferential join condition to be considered when building
     * automatically the query
     *
     * @param from_tbl  DbbSqlTable indicating the primary key table
     * @param from_col  String indicating the primary key table column
     * @param to_tbl    DbbSqlTable indicating the foreign key table
     * @param to_col    String indicating the foreign key table column
     */
    public void addJoinCondition(
            DbbSqlTable from_tbl,
            String from_col,
            DbbSqlTable to_tbl,
            String to_col,
            DbbSqlTable start_tbl,
            DbbSqlTable end_tbl) {

        if (from_col == null || from_col.trim().equals(""))
            return;
        if (to_col == null || to_col.trim().equals(""))
            return;
        if (from_tbl == null || to_tbl == null)
            return;

        from_tbl.completeTable(session_);
        to_tbl.completeTable(session_);

        /* Create a preferred key. */
        DbbSqlKey new_key = new DbbSqlKey(null, from_tbl, from_col, to_tbl, to_col);

        start_tbl.completeTable(session_);
        end_tbl.completeTable(session_);

        new_key.setPreferredKeyFlag(start_tbl, end_tbl);

        joiner_.addDbbSqlKey(new_key);
    }

    /**
     * Add a left outer join condition to be considered when building
     * automatically the query.
     * if the col arrays are not of equal length, nothing happens.
     *
     * @param from_tbl  DbbSqlTable indicating the primary key table
     * @param from_col  String indicating the primary key table column
     * @param to_tbl    DbbSqlTable indicating the foreign key table
     * @param to_col    String indicating the foreign key table column
     */
    public void addCompositeJoinCondition(
            DbbSqlTable from_tbl,
            String from_cols[],
            DbbSqlTable to_tbl,
            String to_cols[],
            boolean leftOutsideJoin) {

        if (from_cols == null || from_cols.length == 0)
            return;
        if (to_cols == null || to_cols.length == 0)
            return;
        if (from_tbl == null || to_tbl == null)
            return;
        if (from_cols.length != to_cols.length) {
            return;
        }

        from_tbl.completeTable(session_);
        to_tbl.completeTable(session_);

        /* Create a DbbSqlLeftOuterJoin key. */
        //DbbSqlKey new_key = new DbbSqlLeftOuterJoinKey( null, from_tbl, from_col,
        //                   to_tbl, to_col);
        DbbSqlKey new_key =
            new DbbSqlCompositeKey(
                null,
                from_tbl,
                from_cols,
                to_tbl,
                to_cols,
                leftOutsideJoin);
        joiner_.addDbbSqlKey(new_key);
    }

    /**
     * Build an SQL query based on the default chunks as well as sume
     * new chunks passed as parameters. Note that default chunks take
     * precedence (they are processed first).
     *
     * @param columns   List of chunks: columns to retrieve
     * @param rows      List of chunks: rows to retrieve
     * @param rows      List of chunks: sort columns
     * @param group     List of chunks: "group by" columns
     * @param isCount   Boolean flag to indicate if it is only a sql count statement
     * @exception SQLException if a jgbc call throws an excpetion
     */
    public String sqlBuildQuery(DbbSqlQueryChunks chunks, boolean isCount)
        throws SQLException {
stdlog_.debug("building query");
        /* Initialize table and clause lists */
        tablesClear();
        whereClausesClear();

        String sql = "";
        String sqlSELECT = "";
        String sqlFROM = "";
        String sqlWHERE = "";
        String sqlSORT = "";
        String sqlGROUP = "";

        /* Add defaults chunks */
        chunks.columns = concatenate(defColumns_, chunks.columns);
        chunks.rows = concatenate(defRows_, chunks.rows);
        chunks.sort = concatenate(defSort_, chunks.sort);
        chunks.group = concatenate(defGroup_, chunks.group);
		stdlog_.debug("added default chunks");

        // @author amchavan, 10-Jun-2002: commented out
        // if(stdlog_.isDebugEnabled())
        //     dumpchunks("buildSQL", chunks.columns, chunks.rows,
        //         chunks.sort, chunks.group);

        /* Precompute list of needed tables. More (join) tables might be
         * added later
         */
        tablesCompute(chunks.columns, chunks.rows, chunks.sort, chunks.group);

        // @author amchavan, 10-Jun-2002: commented out
        // if(stdlog_.isDebugEnabled())
        //   dumptables("Precomputed tables");

        /* The SELECT, SORT and GROUP BY clauses depend on the default
         * chunks and those chunks activated by the user
         */
        sqlSELECT = sqlBuildSELECT(chunks.columns, isCount);

        if (isCount) {
            sqlSORT = "";
            sqlGROUP = "";
        } else {
            sqlSORT = sqlBuildSORT(chunks.sort);
            sqlGROUP = sqlBuildGROUP(chunks.group);
        }

        /* Extract all needed WHERE clauses from the widgets in the row
         * selection panel
         */
        computeRowSelectionClauses(chunks.rows);

        /* Now compute all WHERE join clauses, based on the list of all
         * needed tables
         */
        computeTableJoinClauses();

        // @author amchavan, 10-Jun-2002: commented out
        //if(stdlog_.isDebugEnabled())
        //    dumptables("Join tables");

        sqlWHERE = sqlBuildWHERE();

        /* The FROM clause lists all tables about which we found out so
         * far: default chunks, chunks activated by the user, join
         * tables
         */
        sqlFROM = sqlBuildFROM();

        sql =
            sqlSELECT
                + "\n"
                + sqlFROM
                + "\n"
                + sqlWHERE
                + "\n"
                + sqlSORT
                + "\n"
                + sqlGROUP;

			stdlog_.debug("built sql");

        return sql;
    }

    /**
     * Create and start a thread to get the result set and to process it
     * into an Object array
     *
     * @param sql       Sql statement
     * @exception SQLException if a jdbc call throws an excpetion
     */
    public DbbQuery startQuery(DbbSqlQueryChunks chunks) throws SQLException {

        String sql = this.sqlBuildQuery(chunks, false);
		stdlog_.debug(sql);
        /* Create and start the thread. */
        return execSql(sql);
    }

    /**
     * Calculate the number of rows in a result sert
     *
     * @param sql       Sql statement
     * @exception SQLException if a jdbc call throws an excpetion
     */

    public int execSqlCount(DbbSqlQueryChunks chunks) throws SQLException {
        String sql = this.sqlBuildQuery(chunks, true);

        int count = 0;
		stdlog_.debug(sql);
        /* Open the result set (database cursor) */
        ResultSet rs = session_.execSql(sql);

        while (rs.next()) {
            count = rs.getInt(1);
        }
        rs.close();
        session_.releaseCurrentConnection();
        return count;
    }

    /**
     * Method used by the thread itself to retrieve a result set from
     * the session and to process it into a DbbSqlQueryResult data
     * structure
     *
     * @param sql       Sql statement
     * @exception SQLException if a jdbc call throws an exception
     */
    public DbbQuery execSql(String sql) throws SQLException {
        // @author amchavan, 10-Jun-2002: commented out
        // stdlog_.debug("Getting a result set cursor");

        /* Open the result set (database cursor) */
        ResultSet rs = session_.execSql(sql);
        // @author amchavan, 10-Jun-2002: commented out
        // stdlog_.debug("Processing a result set");

        return new DbbQueryImpl(rs, session_);
    }

    //----------------------------------------
    // Utility methods
    //----------------------------------------

    /**
     *  @return A new array of chunks resulting from concatenating
     *          (appending) array b to array a. Both a and b may be
     *          'null', in which case they are considered as empty
     *          arrays (i.e. zero length)
     */
    static DbbSqlChunk[] concatenate(DbbSqlChunk[] a, DbbSqlChunk[] b) {

        // Cope with null arguments
        if (a == null)
            a = new DbbSqlChunk[0];
        if (b == null)
            b = new DbbSqlChunk[0];

        // Create return array
        int la = a.length;
        int lb = b.length;
        DbbSqlChunk[] ret = new DbbSqlChunk[la + lb];

        // Do the concatenation of a and b into ret
        for (int i = 0; i < a.length; i++)
            ret[i] = a[i];

        for (int i = 0; i < b.length; i++)
            ret[i + la] = b[i];

        return (ret);
    }

    /**
     *  @return A new array of chunks resulting from concatenating
     *          (appending) array b to array a. Both a and b may be
     *          'null', in which case they are considered as empty
     *          arrays (i.e. zero length).
     */
    static DbbSqlChunk[] append(DbbSqlChunk[] a, DbbSqlChunk new_chunk) {

        // Cope with null arguments
        if (a == null)
            a = new DbbSqlChunk[0];
        if (new_chunk == null)
            return a;

        // Create return array
        int la = a.length;
        DbbSqlChunk[] ret = new DbbSqlChunk[la + 1];

        // Do the concatenation of a and b into ret
        for (int i = 0; i < la; i++)
            ret[i] = a[i];

        ret[la] = new_chunk;

        return (ret);
    }

    //----------------------------------------
    // DEBUG
    //----------------------------------------

    static void dumpchunks(
        String header,
        DbbSqlChunk[] columns,
        DbbSqlChunk[] rows,
        DbbSqlChunk[] sort,
        DbbSqlChunk[] group) {

        String sep = "\n----------------------------------------\n";
        stdlog_.debug(sep + header + sep);

        stdlog_.debug("\nCOLUMNS:");
        for (int i = 0; i < columns.length; i++)
            stdlog_.debug(columns[i].toString());

        stdlog_.debug("\nROWS:");
        for (int i = 0; i < rows.length; i++)
            stdlog_.debug(rows[i].toString());

        stdlog_.debug("\nSORT:");
        for (int i = 0; i < sort.length; i++)
            stdlog_.debug(sort[i].toString());

        stdlog_.debug("\nGROUP:");
        for (int i = 0; i < group.length; i++)
            stdlog_.debug(group[i].toString());

        stdlog_.debug("\n");
    }

    private void dumptables(String header) {

        int len = tables_.size();

        String sep = "\n----------------------------------------\n";
        stdlog_.debug(sep + header + sep);

        for (int i = 0; i < len; i++) {
            DbbSqlTable t = (DbbSqlTable)tables_.elementAt(i);
            stdlog_.debug("    <" + t.get() + ">");
        }

        stdlog_.debug("\n");
    }

    private void dumpWhereClauses(String header) {

        int len = whereClauses_.size();

        String sep = "\n----------------------------------------\n";
        stdlog_.debug(sep + header + sep);

        for (int i = 0; i < len; i++)
            stdlog_.debug("    " + whereClauses_.elementAt(i));

        stdlog_.debug("\n");
    }

	public void setNoExec(String noExecMode) throws SQLException {
		if (noExecMode == null ) throw new IllegalArgumentException(MsgManager.errContract("noExecMode is null"));
		if (noExecMode.compareTo("on")!=0 && noExecMode.compareTo("off")!=0) throw new IllegalArgumentException(MsgManager.errContract("noExecMode unexpected value " + noExecMode));
		stdlog_.debug("setting sql noExec " + noExecMode);
		String sql = "set noexec " + noExecMode;
		session_.execSqlNoResulset(sql);
	}
	
	public void execSqlNoResulset(String sql) throws SQLException {
		if (sql == null ) throw new IllegalArgumentException(MsgManager.errContract("sql is null"));
		if (sql.length() == 0 ) throw new IllegalArgumentException(MsgManager.errContract("sql is null"));
		session_.execSqlNoResulset(sql);
	}
}

