package org.eso.ohs.core.dbb.sql;



/** 
 * This is a member, used in a DbbSqlChunkGroup for building composite
 * expressions in the WHERE clause of an SQL statement. 
 * It is constructed as for a normal DbbSqlChunk, but includes a reference
 * to a DbbSqlChunkGroup object.
 *
 * Instead of returning a String in the getSelectClause() method, it
 * calls getExpressionClause() in the DbbSqlChunkGroup, which uses the data of
 * the entire chunk group to build a SQL expression. As each
 * DbbSqlChunk is called with getSelectClause(), the DbbSqlChunkGroup object
 * ensures that the expression is returned through each one of its
 * DbbSqlChunkGroupMember objects.
 *
 * It the responsibility of the code collecting the chunks to ensure that
 * the same expression is not entered multiple time (though this does not
 * affect the resulting SQL statement.
 *
 * Through this mechanism OR statements and nested conditions may be built
 * into the DBB.
 *
 * @author Nick Kornweibel for ESO
 */

public class DbbSqlChunkGroupMember extends DbbSqlChunk {

    protected DbbSqlChunkGroup grp_ = null;

    //--------------------------------------------------
    // Constructors
    //--------------------------------------------------

    /**
     * Create an SQL chunk setting the table, the column, the column
     * label, data type and the SQL operator.<br> This constructor
     * should be used mostly for row selection chunks.
     * 
     * @param table	Database table.
     * @param column	Database column name.
     * @param label	Column label.
     * @param op		SQL default operator.  
     * @param data_type	Database column type (see DbbSqlChunk.datatype_).
     */
    public DbbSqlChunkGroupMember(DbbSqlTable table,
		       String column,
		       String label,
		       DbbSqlOperator op,
		       Class data_type) {
        super(table, column, label, op, data_type);
    }

  /**
     * Create an SQL chunk setting the table, the column, the column
     * label, data type and the SQL operator.<br> This constructor
     * should be used mostly for row selection chunks.
     * 
     * @param table	Database table.
     * @param column	Database column name.
     * @param label	Column label.
     */
    public DbbSqlChunkGroupMember(DbbSqlTable table,
		       String column,
		       String label) {
        super(table, column, label);
    }

  /**
     * Create an SQL chunk setting the table, the column, the column
     * label, data type and the SQL operator.<br> This constructor
     * should be used mostly for row selection chunks.
     * 
     * @param table	Database table.
     * @param column	Database column name.
     */
    public DbbSqlChunkGroupMember(DbbSqlTable table,
		       String column) {
        super(table, column);
    }


    /** 
     * Create an SQL chunk setting the table, the column, data type and
     * the SQL operator.<br> This constructor should be used mostly for
     * row selection chunks.
     * 
     * @param table	Database table.
     * @param column	Database column name.
     * @param op		SQL default operator.  
     * @param data_type	Database column type (see DbbSqlChunk.datatype_).
     */
    public DbbSqlChunkGroupMember(DbbSqlTable table,
		       String column,
		       DbbSqlOperator op,
		       Class data_type) {
    
        super(table, column, op, data_type);
    }
  

    /**
     * Create an SQL chunk setting the table, the column and the
     * database column type.<br> This constructor should be used mostly
     * for column selection chunks.
     * 
     * @param table	Database table.
     * @param column	Database column name.  
     * @param data_type	Database column type (see DbbSqlChunk.datatype_).
     */
    public DbbSqlChunkGroupMember(DbbSqlTable table,
		       String column,
		       Class data_type) {

	super(table, column, data_type);
    }


    /**
     * Create an SQL chunk setting the table, the column and the
     * database column type.<br> This constructor should be used mostly
     * for column selection chunks.
     * 
     * @param table	Database table.
     * @param column	Database column name.  
     * @param label	Column label.  
     * @param data_type	Database column type (see DbbSqlChunk.datatype_).
     */
    public DbbSqlChunkGroupMember(DbbSqlTable table,
		       String column,
		       String label,
		       Class data_type) {

	super(table, column, label, data_type);
    }


    /** 
     * Create an SQL chunk setting the table, the column and the SQL
     * operator.<br> This constructor should be used mostly for row
     * selection chunks.
     * 
     * @param table	Database table.
     * @param column	Database column name.
     * @param op		SQL default operator.  
     */
    public DbbSqlChunkGroupMember(DbbSqlTable table,
		       String column,
		       DbbSqlOperator op) {
    
	super(table, column, op);
    }
  

    /** 
     * Create an SQL chunk setting the table, the column and the SQL
     * operator.<br> This constructor should be used mostly for row
     * selection chunks.
     * 
     * @param table	Database table.
     * @param column	Database column name.
     * @param label	Column label.  
     * @param op		SQL default operator. 
     *
     * <br>ChangeLog: A M Chavan, 18-Feb-2000
     */

    public DbbSqlChunkGroupMember(DbbSqlTable table,
		       String column,
		       String label,
		       DbbSqlOperator op) {
    
	super(table, column, label, op);
    }

    /**
     * Get a select clause from the chunk group.
     * The group will return the expression, derived from all chunks,
     * and return it through this chunk.
     *
     * @return String containing the sql clause.
     */

    public String getSelectClause() {
        String rc = null;
        if (grp_ == null) {
            rc = super.getSelectClause();
        } else {
            rc = grp_.getExpressionClause(this);
        }
        return rc;
    }

    /**
     * Set the DbbSqlChunkGroup of which this chunk is a member.
     */

    public void setSqlChunkGroup(DbbSqlChunkGroup g) {
        grp_ = g;
    }

    /**
     * @return the group of which this chunk is a member
     */

    public DbbSqlChunkGroup getSqlChunkGroup() {
        return grp_;
    }
}
