package org.eso.ohs.core.utilities;

/**
 * Basic implementation of the GTreeModel interface.
 *
 * @author Nick Kornweibel for ESO.
 */

import java.util.Enumeration;
import java.util.Vector;


public class DefaultGTreeModel implements GTreeModel {

    protected GTreeNode root_;

    /**
     * Create a DefaultGTreeModel with root set as the root node.
     */
    public DefaultGTreeModel(GTreeNode root) {
        root_ = root;
    }

    /**
     * @return the root of the tree
     */
    public GTreeNode getRoot() {
        return root_;
    }

    /**
     * @return true if node is a leaf, else false.
     */
    public boolean isLeaf(GTreeNode node) {
        return node.isLeaf();
    }

    /**
     * @return the child of node with the name childName, or null if not found.
     */
    public GTreeNode getChild(GTreeNode node, String childName) {
        return node.getChild(childName);
    }

    /**
     * @return the child of node with the name child.getName(),
     * or null if not found.
     */
    public GTreeNode getChild(GTreeNode node, GTreeNode child) {
        return node.getChild(child);
    }

    /**
     * @return the number of children of node.
     */
    public int getChildCount(GTreeNode node) {
        return node.getChildCount();
    }

    /**
     * @return an Enumeration of node's children.
     */
    public Enumeration getChildren(GTreeNode node) {
        return node.getChildren();
    }

    /**
     * @return an array of GTreeNode objects,
     * with the 0th element being root and
     * the last element being aNode.
     * If there is a cycle in the Tree map, this method may not terminate.
     */
    public GTreeNode[] getPathToRoot(GTreeNode aNode) {

        Vector path = new Vector();

        while (aNode != null) {
            path.addElement(aNode);
            if (aNode.equals(getRoot())) {
                break;
            }
            aNode = aNode.getParent();
        }

        GTreeNode[] rc = new GTreeNode[path.size()];
        for (int i = 0; i < rc.length; i++) {
            rc[i] = (GTreeNode)path.elementAt(rc.length-(i+1));
        }

        return rc;
    }
 
    /**
     * Add child as a child or aNode (no checking for duplicates).
     * Note: no checking for duplicates!
     */
    public void addChild(GTreeNode aNode, GTreeNode child) {
        aNode.addChild(child);
    }

    /**
     * Add child as a child or aNode (no checking for duplicates).
     * Note: no checking for duplicates!
     */
    public boolean addPath(GTreePath child) {
        return addPath(getRoot(), child);
    }

    /**
     * Add a tree path to the root.
     * @return false if the first node on the path is not the same 
     * as the tree root
     */
    public boolean addPath(GTreeNode aNode, GTreePath child) {
        boolean rc;
        GTreeNode[] path = child.getPath();
        int startAppend;
        if (!aNode.getName().equals(path[0].getName())) {
            rc = false;
        } else {
            for (startAppend = 1; startAppend < path.length; startAppend++) {
                GTreeNode n = aNode.getChild(path[startAppend]);
                if (n == null) {
                    break;
                } else {
                    aNode = n;
                }
            }
            for (int i = startAppend; i < path.length; i++) {
                aNode.addChild(path[i]);
                aNode = path[i];
            }
            rc = true;
        }
        return rc;
    }

    /**
     * @return (and remove) child from aNode, or null if not found.
     */
    public GTreeNode removeChild(GTreeNode aNode, GTreeNode child) {
        return aNode.removeChild(child);
    }
   
    /**
     * @return (and remove) the first child of aNode with the name childName.
     */
    public GTreeNode removeChild(GTreeNode aNode, String childName) {
        return aNode.removeChild(childName);
    }

    /**
     * @return the Node specified by the path, or null if not found.
     * path[0] must be the root node and path[path.length-1]
     * is the node required. Tree nodes
     * are traversed using the names of the path nodes.
     */
    public GTreeNode getNode(GTreeNode[] path) {
        GTreeNode rc = null;
        if (path != null) {
            GTreeNode node = getRoot();
            if (node.equals(path[0])) {
                for (int i = 1; node != null && i < path.length; i++) {
                    node = node.getChild(path[i]);
                }
                rc = node;
            } else {
                rc = null;
            }
        }
        return rc;
    }

    /**
     * @return the Node specified by the path, or null if not found.
     */

    public GTreeNode getNode(GTreePath path) {
        return getNode(path.getPath());
    }


    /**
     * @return the String value of the node specified by path, or null
     * if the node was not found.
     */
    public String getValue(GTreePath path) {
        GTreeNode node = getNode(path);
        String rc = null;
        if (node != null) {
            rc = node.getValue();
        }
        return rc;
    }

    /**
     * @return the value of the node specified by path, or null
     * if the node was not found.
     */
    public Object get(GTreePath path) {
        GTreeNode node = getNode(path);
        Object rc = null;
        if (node != null) {
            rc = node.get();
        }
        return rc;
    }

    /**
     * Recursive method for finding tree nodes by name.
     */
    public GTreeNode findNode(String name) {
        return findNode(getRoot(), name);
    }

    public GTreeNode findNode(GTreeNode n) {
        return findNode(getRoot(), n.getName());
    }

    protected GTreeNode findNode(GTreeNode node, String n) {
       GTreeNode rc = null;
       if (node.getName().equals(n)) {
           rc = node;
       } else {
           for (Enumeration e = node.getChildren(); e.hasMoreElements(); ) {
               rc = findNode((GTreeNode)e.nextElement(), n);
               if (rc != null) {
                   break;
               }
           }
       }
       return rc;
    }

    /**
     * @return a string representation of the tree in depth first traversal.
     */
    public String toString() {
        return toString(getRoot());
    }

    /**
     * @return a String representation of the tree (from n down) in depth
     * first traversal.
     */
    public String toString(GTreeNode n) {
        StringBuffer rc = new StringBuffer();
        printTree(n, 0, rc);
        return rc.toString();
    }

    /**
     * Recursive method for printing tree nodes.
     */
    protected void printTree(GTreeNode n, int depth, StringBuffer rc) {
       pad(rc, depth);
       rc.append(n.getName()+"("+n.getValue()+")\n");
       for (Enumeration e = n.getChildren(); e.hasMoreElements(); ) {
           printTree((GTreeNode)e.nextElement(), depth+1, rc);
       }
    }

    protected void pad(StringBuffer rc, int i) {
        while (i-- > 0) {
            rc.append("  ");
        }
    }
}
