package org.eso.ohs.core.utilities;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.util.Vector;

import org.apache.log4j.Logger;

/**
 *    Run external command in an OHS application.
 */
public class RunCommand {

	/** Command string, broken up into components */
	private String[] commandTokenList_ = null;
	;

	/** Command string, single. */
	private String simpleCommand_ = null;
	;

	private StringBuffer outString;
	private StringBuffer errString;

	private static Logger stdlog_ = Logger.getLogger(RunCommand.class);

	/**
	 * @param cmd  Command string to be executed and its parameters. 
	 *             For instance, <P>
	 *             <code>
	 *             String[] cmd = { "/bin/ls", "-l" );<BR>
	 *             RunCommand cmd = new RunCommand( cmd );
	 *             </code>
	 *             <P>
	 *             This constructor must be used when one or more of
	 *             the command parameters need to be quoted.
	 *             The & and ; characters are considered as command separators
	 *
	 * @throws IllegalArgumentException If parameter <em>cmd</em> is null.
	 */
	public RunCommand(String[] cmd) throws IllegalArgumentException {
		if (cmd == null)
			throw new IllegalArgumentException("null argument");

		commandTokenList_ = cmd;
		outString = new StringBuffer();
		errString = new StringBuffer();
	}

	/**
	 * @param cmd Command string to be executed. For instance, <BR>
	 *            <code>bin/my-script.sh</code>
	 *
	 * @throws IllegalArgumentException If parameter <em>cmd</em> is null.
	 */
	public RunCommand(String cmd) throws IllegalArgumentException {
		if (cmd == null)
			throw new IllegalArgumentException("null argument");

		simpleCommand_ = cmd;
		outString = new StringBuffer();
		errString = new StringBuffer();
	}

	/**
	 * The output is kept only when 
	 *  executeCommand() is called.
	 * @see executeCommand()
	 * @return output result of the last execution 
	 */
	public String getOutput() {
		return outString.toString();
	}
	
	
	/**
	 * @param substring string to be matched
	 * @return boolean true if substring present
	 */
	public boolean outputContains(String substring) {
		boolean result = false;
		if (substring != null && outString.indexOf(substring)>= 0) {
	        result = true;
	    }
	
        return result;
	}

	/**
	 * @return Human-readable form of the command string
	 */
	public String getCommand() {

		if (simpleCommand_ != null)
			return (simpleCommand_);

		StringBuffer t = new StringBuffer();

		for (int i = 0; i < commandTokenList_.length; i++) {
			t.append('\'');
			t.append(commandTokenList_[i]);
			t.append('\'');
			t.append(' ');
		}
		return t.toString();
	}

	/**
	 * @return error result of the last execution
	 */
	public String getError() {
		return errString.toString();
	}

	/**
	 * Executes the command.
	 *  Standard  error can be retrieved using getError().
	 *  @param out output stream  (the error stream is saved into a string)
	 * @throws CommandExecutionException Child process terminated with errors, or
	 * threading problems with subprocess. 
	 * @throws IOException Could not read from
	 * child process' stdout and stderr streams.
	 */
	public void executeCmd(OutputStream out)
		throws CommandExecutionException, IOException {
		stdlog_.debug("Executing: '" + getCommand() + "'");

		String s = null;
		Process p = null;
		if (simpleCommand_ != null) {

			p = Runtime.getRuntime().exec(simpleCommand_);
			checkProcessStream(out, p);
		} else {
			int nextGlobalPosition = 0;
			int startNextCmd = 0;
			int currentCmdListLength = 0;

			while (commandTokenList_.length > nextGlobalPosition) {

				Vector commandVector = new Vector();
				for (int i = startNextCmd; i < commandTokenList_.length; i++) {

					currentCmdListLength = i - startNextCmd;
					nextGlobalPosition = i + 1;
					if (commandTokenList_[i].equals("&")
						|| commandTokenList_[i].equals(";")) {
						startNextCmd = i + 1;
						break;
					}
					commandVector.addElement(commandTokenList_[i]);
				}
				String[] singleCommandList =
					(String[]) commandVector.toArray(
						new String[currentCmdListLength]);
				stdlog_.debug(
					"execute "
						+ TextUtils.arrayToString(singleCommandList, " ", false));
				p = Runtime.getRuntime().exec(singleCommandList);
				checkProcessStream(out, p);

			}

		}
	}

	private void checkProcessStream(
		OutputStream out,
		Process p)
		throws CommandExecutionException, IOException {

		BufferedInputStream stdOutput = null;
		BufferedReader stdError = null;
		String s;
		try {
			// For some bizarre naming convention, in order to read a
			// process' stdout stream one must call method
			// Process.getInputStream()... AMC & FS, 21-Jan-2003
			//-------------------------------------------------------
			stdOutput = new BufferedInputStream(p.getInputStream());

			stdError =
				new BufferedReader(new InputStreamReader(p.getErrorStream()));

			// Read standard output stream of the command:
			// if any, log to a file
			//--------------------------------------------------

			byte[] buff = new byte[2048];
			int bytesRead;
			//			   Simple read/write loop.
			while (-1 != (bytesRead = stdOutput.read(buff, 0, buff.length))) {
				out.write(buff, 0, bytesRead);
			}

			// IF return code was not zero
			// OR the standard output stream is not empty
			// THEN log the error text (if any) 
			//      throw an exception
			//--------------------------------------------------

			while ((s = stdError.readLine()) != null) {
				errString.append(s);
				errString.append('\n');
			}
			stdError.close();
			int retcode = p.waitFor();
			retcode = p.exitValue();

			if (retcode != 0) {
				stdlog_.error(
					"Exit value: '"
						+ retcode
						+ "'; should be 0: "
						+ errString.toString());
				throw new CommandExecutionException("\nCommand:"+getCommand()+"\n**********\n "+errString.toString());
			}

			if (errString.length() != 0) {
				stdlog_.error(
					"Standard error message not empty:'"
						+ errString
						+ "'. Throw exception");
				throw new CommandExecutionException("\nCommand:"+getCommand()+"\n**********\n "+errString.toString());
			}

		} catch (java.lang.IllegalThreadStateException e) {
			stdlog_.error(e);
			throw new CommandExecutionException(e.getMessage());
		} catch (java.lang.InterruptedException e) {
			stdlog_.error(e);
			throw new CommandExecutionException(e.getMessage());
		} catch (IOException e) {
			stdlog_.error(e);
			// errString.append(e.getMessage());  ==> useless
			throw e;
		} finally {

			if (out != null) {
				out.flush();
				out.close();
			}
			if (stdOutput != null) {
				stdOutput.close();
			}
			if (stdError != null) {
				stdError.close();
			}
		}
	}

	/**
		 * Executes the command.
		 *
		 * Standard output and standard error can be retrieved using getError() and
		 * getOutput().
		 * 
		 * @throws CommandExecutionException Child process terminated with errors, or
		 * threading problems with subprocess.
		 * @throws IOException Could not read from
		 * child process' stdout and stderr streams.
		 */
	public void executeCmd() throws CommandExecutionException, IOException {
		stdlog_.debug("Executing: '" + getCommand() + "'");
		String s = null;
		outString = new StringBuffer();
		errString = new StringBuffer();
		Process p;

		if (simpleCommand_ != null) {
			p = Runtime.getRuntime().exec(simpleCommand_);
			checkProcessStream(p);
		} else {

			int nextGlobalPosition = 0;
			int startNextCmd = 0;
			int currentCmdListLength = 0;

			while (commandTokenList_.length > nextGlobalPosition) {

				Vector commandVector = new Vector();
				for (int i = startNextCmd; i < commandTokenList_.length; i++) {

					currentCmdListLength = i - startNextCmd;
					nextGlobalPosition = i + 1;
					if (commandTokenList_[i].equals("&")
						|| commandTokenList_[i].equals(";")) {
						startNextCmd = i + 1;
						break;
					}
					commandVector.addElement(commandTokenList_[i]);
				}
				String[] singleCommandList =
					(String[]) commandVector.toArray(
						new String[currentCmdListLength]);
				stdlog_.debug(
					"execute "
						+ TextUtils.arrayToString(singleCommandList, " ", false));
				p = Runtime.getRuntime().exec(singleCommandList);
				checkProcessStream(p);

			}
		}
	}

	private void checkProcessStream(Process p)
		throws CommandExecutionException, IOException {
		String s;
		try {
			// For some bizarre naming convention, in order to read a
			// process' stdout stream one must call method
			// Process.getInputStream()... AMC & FS, 21-Jan-2003
			//-------------------------------------------------------
			BufferedReader stdOutput =
				new BufferedReader(new InputStreamReader(p.getInputStream()));

			BufferedReader stdError =
				new BufferedReader(new InputStreamReader(p.getErrorStream()));

			// Read standard output stream of the command:
			// if any, log it as debug
			//--------------------------------------------------
			s = stdOutput.readLine();
			while (s != null) {
				outString.append(s);
				s = stdOutput.readLine();
				if (s != null) {
				 outString.append('\n');
				}
			}
			if (outString.length() != 0) {
				stdlog_.debug("Standard output:");
				stdlog_.debug(outString);
			}

			// IF return code was not zero
			// OR the standard output stream is not empty
			// THEN log the error text (if any) 
			//      throw an exception
			//--------------------------------------------------

			while ((s = stdError.readLine()) != null) {
				errString.append(s);
				errString.append('\n');
			}

			int retcode = p.waitFor();
			retcode = p.exitValue();

			if (retcode != 0) {
				stdlog_.error("Exit value: '" + retcode + "'; should be 0");
				throw new CommandExecutionException("\nCommand:"+getCommand()+"\n**********\n "+errString.toString());
			}

			if (errString.length() != 0) {
				stdlog_.error(
					"Standard error message not empty:'"
						+ errString
						+ "'. Throw exception");
				throw new CommandExecutionException("\nCommand:"+getCommand()+"\n**********\n "+errString.toString());
			}

		} catch (java.lang.IllegalThreadStateException e) {
			stdlog_.error(e);
			throw new CommandExecutionException(e.getMessage());
		} catch (java.lang.InterruptedException e) {
			stdlog_.error(e);
			throw new CommandExecutionException(e.getMessage());
		} catch (IOException e) {
			stdlog_.error(e);
			// errString.append(e.getMessage());  ==> useless
			throw e;
		}
	}
}
