package org.eso.ohs.dfs;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.ohs.core.docview.datatrans.AdaptorFactory;
import org.eso.ohs.core.docview.datatrans.ObjectAdaptor;
import org.eso.ohs.core.utilities.Convert;
import org.eso.ohs.core.utilities.ReflectionException;
import org.eso.ohs.core.utilities.TextUtils;
import org.eso.ohs.instruments.Instrument;
import org.eso.ohs.instruments.InstrumentConstraintConfig;
import org.eso.ohs.instruments.InstrumentList;
import org.eso.ohs.instruments.Parameter;
import org.eso.ohs.instruments.TemplateSignature;
import org.eso.ohs.instruments.TemplateSignatureFactory;

/**
 * 
 * 
 * <p>
 *
 *
 *
 * $Id: ObservationBlock.java,v 1.108 2005/09/07 07:59:59 tcanavan Exp $
 *
 *
 * <br>
 * @author Bob Kemp
 */

public class ObservationBlock extends CalibrationBlock {

    private static Logger stdlog_ = Logger.getLogger(ObservationBlock.class);

    /** Version id for object serialization */
    static final long serialVersionUID = -3964774802755871193L;

    /**
     *  Associated Target object
     */
    private transient Target target_;

    /**
     *  Associated ConstraintSet object
     */
    private transient ConstraintSet constraintSet_;

    /**
     *  Associated observation time intervals
     */
    private TimeInterval[] timeIntervals_;

    /**
         *  Associated observation time intervals
         */
    private TimeInterval[] stIntervals_;

    /**
     *  Signature of the acquisition template associated to this
     *  observation block. If no acquisition template is associated,
     *  its value a null handle (the empty string).
     */
    private TemplateSignature templateSignature_;

    /**
     *  User-defined text to explain calibration requirements for OB.
     */
    private String calibrationRequirements_;

    /**
     *  Observing Grade
     */
    private String obsGrade_ = "";

    private FindingChart[] findingCharts;
    
    private EphemerisFile  ephemerisFile;
    
    
    /**
     * Default constructor, mainly used by deserialization code.
     */
    ObservationBlock() {
        //org.eso.ohs.utilities.DebugUtils.stackTrace();
    }

    /**
     * Returns the value of the database 'type' column for this object
     * 
     * @return  database 'type'-column value
     */
    public String getType() {
        return OBSBLK_TYPE;
    }

    public void setType(String s) {
    }

    /**
     * Set new value for timeIntervals_ property
     *
     * @param newValue the new value for the property
     */
    public void setTimeIntervals(TimeInterval[] newValue) {
        timeIntervals_ = newValue;
        firePropertyChangeEvent();
    }

    /**
     * Get value for timeIntervals_ property
     *
     * @return the current value of the property
     */
    public TimeInterval[] getTimeIntervals() {
        if(timeIntervals_ == null)
             return new TimeInterval[0];
        return timeIntervals_;
    }

    /**
     * Set new value for target_ property
     *
     * @param newValue the new value for the property
     */
    public void setTarget(Target newValue) {
        if (target_ != null) {
            target_.removePropertyChangeListener(eventRedirector_);
        }
        target_ = newValue;
        if (target_ != null) {
            target_.addPropertyChangeListener(eventRedirector_);
        }
        firePropertyChangeEvent();
    }

    /**
      * Return the first template signature in the list of templates
      * @return TemplateSignature
      */

    public TemplateSignature firstTS() {
        TemplateSignature ts = getTemplateSignature();
        if (ts != null) {
            return ts;
        }
        return null;
    }

    /**
     * Get value for target_ property
     *
     * @return the current value of the property
     */
    public Target getTarget() {
        return target_;
    }

    /**
     * Set new value for calibrationRequirements_ property
     *
     * @param newValue the new value for the property
     */
    public void setCalibrationRequirements(String newValue) {
        calibrationRequirements_ = newValue;
        firePropertyChangeEvent();
    }

    /**
     * Get value for calibrationRequirements_ property
     *
     * @return the current value of the property
     */
    public String getCalibrationRequirements() {
        if ( calibrationRequirements_ == null) 
            return "";
        else
            return calibrationRequirements_;
    }

    /**
     * @see #obsGrade_
     *
     * @param newValue the new value for the property
     */
    public void setObsGrade(String obsGrade) {
        obsGrade_ = obsGrade;
        firePropertyChangeEvent();
    }

    /**
     * @see #obsGrade_
     *
     * @return the current value of the property
     */
    public String getObsGrade() {
        if (obsGrade_ == null)
            return("");
        else
            return obsGrade_;
    }

    /**
     * Set new value for constraintSet_ property
     *
     * @param newValue the new value for the property
     */
    public void setConstraintSet(ConstraintSet newValue) {
        if (constraintSet_ != null) {
            constraintSet_.removePropertyChangeListener(eventRedirector_);
        }
        constraintSet_ = newValue;
        if (constraintSet_ != null) {
            constraintSet_.addPropertyChangeListener(eventRedirector_);
        }
        constraintSet_ = newValue;
        firePropertyChangeEvent();
    }

    /**
     * Get value for constraintSet_ property
     *
     * @return the current value of the property
     */
    public ConstraintSet getConstraintSet() {
        return constraintSet_;
    }

    /**
     * Set new value for templateSignature_ property
     *
     * @param newValue the new value for the property
     */
    public void setTemplateSignature(TemplateSignature newValue) {
        templateSignature_ = newValue;
        firePropertyChangeEvent(templateEvent_);
    }

    /**
     * getTemplateSignature returnes the Signature of the acquisition
     * template associated to this observation block. If no acquisition
     * template is associated, If no acquisition template is associated
     * it returnes null.
     *
     * @param void
     * @return the current value of the property
     * <br>ChangeLog: Toni Schlichting (added a sufficient comment) 15-June-2001
     */
    public TemplateSignature getTemplateSignature() {
        return templateSignature_;
    }

    /**
     * Returns text representing this object in OBD format suitable
     * for passing to the VLT software.
     * @param addHeader specifies if the OBD Header must be included 
     * @return this object in OBD format
     */

    public String toOBD(boolean addHeader, boolean display,long id,String desc) {
        String text = getOBDHeader(desc);

        if (addHeader) {
            text = getOBDHeader(desc) + getOBDBody(id);
        } else {
            text = getOBDBody(id);
        }

        if (getTarget() != null) {
            String tgName = getTarget().getName();

            tgName = tgName.replace(' ', '-').replace('\t', '-');
            text += "\nOBS.TARG.NAME  \"" + tgName + "\"";
        }

        text += "\n";
        text += "\n";

        if (getTemplateSignature() != null) {
            text += toParameterFormat(templateSignature_, 1, display);
        }
        if (getOd() != null) {
            TemplateSignature[] tss = getOd().getSignatures();
            for (int j = 0; j < tss.length; j++) {
                text += toParameterFormat(tss[j], j + 2, display);
            }
        }

        return text;
    }

    public String getOBDParameterValue(Parameter p, boolean display) {
        String value = null;
        if (target_ != null) {

            if (p.samePrefix("TEL.TARG.ALPHA")
                || p.samePrefix("TEL.TARG.RA") ||
                p.samePrefix("TEL.TARG.ADDVELALPHA")) {
                value = Convert.convertToXXMMSS(target_.getRA());

            } else if (
                p.samePrefix("TEL.TARG.DELTA")
                    || p.samePrefix("TEL.TARG.DEC")
                    || p.samePrefix("TEL.TARG.ADDVELDELTA")) {
                value = Convert.convertToXXMMSS(target_.getDeclination());
            } else if (p.samePrefix("TEL.TARG.EQUINOX")) {
                value = target_.getEquinox();
                // XXX drop first character if alphabetic ???
                if (value.length() > 0
                    && Character.isLetter(value.charAt(0))) {
                    value = value.substring(1);
                }
            } else if (p.samePrefix("TEL.TARG.EPOCH")) {
                value = String.valueOf(target_.getEpoch());
            } else if (
                p.samePrefix("TEL.TARG.PROPRA")
                    || p.samePrefix("TEL.TARG.PMA")) {
                value = String.valueOf(target_.getPropRA());
            } else if (
                p.samePrefix("TEL.TARG.PROPDEC")
                    || p.samePrefix("TEL.TARG.PMD")) {
                value = String.valueOf(target_.getProperDeclination());
            } else if (p.samePrefix("TEL.TARG.COLOR")) {
                value = String.valueOf(target_.getColor());
            } else if (p.samePrefix("TEL.TARG.MAG")) {
                value = String.valueOf(target_.getMag());
            } else if (p.samePrefix("TEL.TARG.NAME")) {
                value = target_.getName();
            }
        }
        if (value == null) {
            if (display)
                value = p.toDisplayOBDParameterString();
            else
                value = p.toOBDParameterString();
        } else {
            if (display)
                value = p.toDisplayOBDParameterString(value);
            else
                value = p.toOBDParameterString(value);
        }
        return value;
    }

    /**
            * Get the inetger value at the end of a keyword  
            * @param  keyword Kwyword to search for integer
            * @return Integer value  at the end of a keyword
            */

    private static int getKeywordIndex(String keyword) {
        int index = -1;
        int startIndex = -1;
        int endIndex = -1;

        for (int k = 0; k < keyword.length(); k++) {
            if (Character.isDigit(keyword.charAt(k))) {
                startIndex = k;
            }
        }
        //dIndex = keyword.lastIndexOf('.');
        String idxStr = keyword.substring(startIndex, startIndex + 1);
        try {
            index = new Integer(idxStr).intValue();
        } catch (NumberFormatException e) {
            //stdlog_.error( e );
        }
        return index;
    }

    /**
     * Return the parameter string needed for communication with BOB
     * @return String to send to BOB for execution.
     */

    public String toParameterFormat(
        TemplateSignature ts,
        int seqno,
        boolean display) {
        StringBuffer sb = new StringBuffer();
        sb.append(getOBDTemplateHeader(ts, seqno));
        Parameter[] params = ts.getParamList();
        String vimosDateTimeStr = TextUtils.currentISODateTime();
        for (int j = 0; j < params.length; j++) {
            if (params[j].isFixed()) {
                continue;
            }
            if (params[j].getParameterName().indexOf("INS.ADM") != -1
                && getOd().getInstrument().equals("VIMOS"))
                sb.append(
                    getVIMOSADMParameterValue(
                        params[j],
                        display,
                        vimosDateTimeStr));
            else
                sb.append(getOBDParameterValue(params[j], display));
        }
        sb.append("\n");
        return sb.toString();
    }

    /**
     * Returns the string used to refer to this type of BusinessObject
     * in messages to the user.
     *
     * @return user-oriented name
     */
    public static String getDisplayName() {
        return "ObsBlock";
    }

    /**
     * Returns the filename suffix for files containing this type of object.
     *
     * @return suffix for filename when stored in local filesystem
     */
    public static String getSuffix() {
        return "ob";
    }

    /**
     * Returns an array of the immediate child BusinessObjects for
     * this BO.  
     * 
     * @return an array of business objects referred to by this BO.
     */
    public BusinessObject[] getBusObjChildren() {
        Vector childVec = new Vector();

        if (getTarget() != null) {
            childVec.addElement(getTarget());
        }

        if (getOd() != null) {
            childVec.addElement(getOd());
        }

        if (getConstraintSet() != null) {
            childVec.addElement(getConstraintSet());
        }

        BusinessObject[] children = new BusinessObject[childVec.size()];
        childVec.copyInto(children);

        return children;
    }

    /**
     * Returns a deep copy of this BO, ie recursively copies this BO
     * and all of it's attached BOs.
     * 
     * @return a deep copy of this business object.
     */
    public BusinessObject deepCopy() {
        ObservationBlock ob = null;
        ObservationDescription od = null;
        ConstraintSet cs = null;
        Target tg = null;
        ob = (ObservationBlock) copy();
        if (getOd() != null) {
            od = (ObservationDescription) getOd().deepCopy();
            ob.setOd(od);
        }

        if (getTarget() != null) {
            tg = (Target) getTarget().deepCopy();
            ob.setTarget(tg);
        }

        if (getConstraintSet() != null) {
            cs = (ConstraintSet) getConstraintSet().deepCopy();
            ob.setConstraintSet(cs);
        }

        return ob;
    }

    /**
     * Copies the value of the specified property of this object to
     * the given BusinessObject.
     * 
     * @param propName      property name to copy
     * @param destBO        destination for the copy operation
     */
    public void copyProperty(String propName, BusinessObject destBO)
        throws
            NoSuchMethodException,
            IllegalAccessException,
            InvocationTargetException {
        ObservationBlock ob = (ObservationBlock) destBO;

        if (propName.equals("TemplateSignature")) {
            if (getTemplateSignature() != null) {
                ob.setTemplateSignature(
                    new TemplateSignature(getTemplateSignature()));
            } else {
                ob.setTemplateSignature(null);
            }
            return;
        }

        if (propName.equals("TimeIntervals")) {
            TimeInterval[] oldTIs = getTimeIntervals();

            if (oldTIs != null) {
                TimeInterval[] newTIs = new TimeInterval[oldTIs.length];

                for (int j = 0; j < newTIs.length; j++) {
                    newTIs[j] = new TimeInterval(oldTIs[j]);
                }

                ob.setTimeIntervals(newTIs);
            } else {
                ob.setTimeIntervals(new TimeInterval[] {
                });
            }
            return;
        }

        if (propName.equals("FindingCharts")) {
            FindingChart[] dbFCs = getFindingCharts();
            if (dbFCs != null) {
                FindingChart[] newFCs = new FindingChart[dbFCs.length];
                FindingChart[] localFCs =
                    ((ObservationBlock) destBO).getFindingCharts();
                for (int i = 0; i < newFCs.length; i++) {
                    newFCs[i] = new FindingChart(dbFCs[i]);
                }
                ob.setFindingCharts(newFCs);
            } else {
                ob.setFindingCharts(new FindingChart[] {
                });
            }
            return;
        }
        super.copyProperty(propName, destBO);
    }

    public String toReport(long id) {
        String ra;
        String dec;
        String equinox = target_.getEquinox();
        ObjectAdaptor raAdaptor = AdaptorFactory.getAdaptor(Target.class, "RA");
        ObjectAdaptor decAdaptor =
            AdaptorFactory.getAdaptor(Target.class, "Declination");
        ra = raAdaptor.fromMetaLevelObject(new Integer(target_.getRA()));
        dec =
            decAdaptor.fromMetaLevelObject(
                new Integer(target_.getDeclination()));
        StringBuffer buf = new StringBuffer();
        long obId = id;
        if (getId() >= 0) { // not local
            obId = getDbaseId();
        }
        buf.append(getName());
        buf.append(" ");
        if (getDbaseId() != 0) {
                buf.append("(OB Id="
                        + obId
                        + ") ");
        }
        buf.append(getObsRun().getName() + "\n");
        if (calibrationRequirements_ != null
            && calibrationRequirements_.length() > 0) {
            buf.append("Cal Req.  : ");
            String[] calStrings =
                TextUtils.splitString(calibrationRequirements_, 60);
            for (int i = 0; i < calStrings.length; i++) {
                if (i != 0)
                    buf.append("            ");
                buf.append(calStrings[i] + "\n");
            }
        }
        String usrCom = getUserComments();
        if (usrCom != null && usrCom.length() > 0) {
            buf.append("User Com. : ");
            String[] usrStrings = TextUtils.splitString(usrCom, 60);
            for (int i = 0; i < usrStrings.length; i++) {
                if (i != 0)
                    buf.append("            ");
                buf.append(usrStrings[i] + "\n");
            }

        }
        buf.append("\n");
        buf.append(
            "Target    : "
                + target_.getName()
                + ","
                + " RA: "
                + ra
                + ", "
                + "Dec: "
                + dec
                + ", "
                + "Eq: "
                + equinox
                + "\n");
        String tgCom = getTarget().getUserComments();
        if (tgCom != null && tgCom.length() > 0) {
            buf.append("User Com. : ");
            String[] tgStrings = TextUtils.splitString(tgCom, 60);
            for (int i = 0; i < tgStrings.length; i++) {
                if (i != 0)
                    buf.append("            ");
                buf.append(tgStrings[i] + "\n");
            }
        }
        buf.append("\n");
        buf.append("ObsDesc: " + getOd().getName() + "\n");
        buf.append("\n");
        if (templateSignature_ != null) {
            buf.append(templateSignature_.toReportString());
            buf.append("\n");
        }
        TemplateSignature[] tsfs = getOd().getSignatures();
        for (int i = 0; i < tsfs.length; i++) {
            buf.append(tsfs[i].toReportString());
            buf.append("\n");
        }
        buf.append(getConstraintSet().toReportString(getOd().getInstrument(),fetchIPVersionFromOd()));
        fillTimeInterval(buf,timeIntervals_);

        if (stIntervals_ != null && stIntervals_.length > 0) {
            buf.append("\n");
            buf.append("Sidereal Time Intervals : ");
            for (int i = 0; i < stIntervals_.length; i++) {
                if (i != 0) {
                    buf.append("                          ");
                }
                buf.append(stIntervals_[i].toReportStringSidereal() + "\n");
            }
        }
        if (findingCharts != null && findingCharts.length > 0) {
            buf.append("\n");
            buf.append("FindingCharts : ");
            for (int i = 0; i < findingCharts.length; i++) {
                if (i != 0) {
                    buf.append("                ");
                }
                buf.append(findingCharts[i].getFcName() + "\n");
            }
        }
        buf.append(
            "\n"
                + "====================================================="
                + "\n");
        return buf.toString();
    }

    /**
     * @param buf void
     */
    public static void fillTimeInterval(StringBuffer buf,TimeInterval[] timeIntervals) {
        if (timeIntervals!= null && timeIntervals.length > 0) {
            buf.append("\n");
            buf.append("Time Intervals : ");
            for (int i = 0; i < timeIntervals.length; i++) {
                if (i != 0) {
                    buf.append("                 ");
                }
                buf.append(timeIntervals[i].toReportString() + "\n");
            }
        }
    }

    /**
     * Verifies this block and all attached blocks, returning the
     * concatenation of all error messages.
     * 
     * @return the (possibly empty) concatenation of all error messages.
     */
    public String verify() {
        StringBuffer sb = new StringBuffer();

        sb.append(super.verify());

        if (getTarget() == null) {
            sb.append("\n Target not defined.");
        } else {
            sb.append(getTarget().verify());
        }
        if (getConstraintSet() == null) {
            if ("S".equalsIgnoreCase(getObsRun().getObsMode()))
                sb.append("\n Constraint Set not defined.");
        } else {
            sb.append(getConstraintSet().verify());
        }
        sb.append(verifyBO());
        return sb.toString();
    }

    public String verifyBO() {
        StringBuffer sb = new StringBuffer();
        if (timeIntervals_ != null && timeIntervals_.length > 0) {
            for (int i = 0; i < timeIntervals_.length; i++) {
                String msg = timeIntervals_[i].verify();
                if (msg != null && msg.length() > 0) {
                    int index = i + 1;
                    sb.append("\n Time interval " + index + " is invalid.");
                }
            }
        }
        if (getTemplateSignature() == null) {
            sb.append("\n Acquisition Template not defined.");
        } else {
            int numErrs = 0;
            String[] errs = getTemplateSignature().verify();

            for (int k = 0; k < errs.length; k++) {
                if (!errs[k].equals("OK")) {
                    numErrs++;
                }
            }

            if (numErrs > 0) {
                sb.append("\n");
                if (numErrs == 1) {
                    sb.append(" There is 1 error");
                } else {
                    sb.append(" There are " + numErrs + " errors");
                }
                sb.append(" in the acquisition template.");
            }
        }
        return sb.toString();

    }

    /**
     * MIPS Project:
     * Set the ipversion of all templates in this object. Not, the "conversion"
     * to the new template does not occur until refreshTemplates() is called.
     */
    public void convertIPVersion(float ipv) {
        if (templateSignature_ != null) {
            templateSignature_.setVersion(ipv);
        }
        super.convertIPVersion(ipv);
    }

    /**
     * MIPS Project:
     * Reload the templates (converting as required to a new IP version).
     * The firePropChangeEvent indicates if a property change event should
     * be fired as a result of the change (this will trigger the autosave).
     */
    public void refreshTemplates(boolean firePropChangeEvent)
        throws IOException {
        refreshTemplates(firePropChangeEvent, false);
    }

    public void refreshTemplates(
        boolean firePropChangeEvent,
        boolean fullRefresh)
        throws IOException {

        if (templateSignature_ != null) {
            if (!TemplateSignatureFactory
                .refreshTemplate(templateSignature_, fullRefresh)) {
                setTemplateSignature(null);
            }
        }
        super.refreshTemplates(firePropChangeEvent, fullRefresh);
    }

    /**
     * Read varianle ob details from local storage
     * @param in Stream to capture data from cache
     * @exception IOException IO Exception if record cannot be read from disk
     * @exception ClassNotFoundException class not found exception
     */

    private void readObject(java.io.ObjectInputStream in)
        throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        if (templateSignature_ != null) {
            if (!TemplateSignatureFactory
                .refreshTemplate(templateSignature_)) {
                // throw an exception if the Instrument does not even exist.
                if (templateSignature_.getInstrument() == null) {
                    throw new IOException(
                        "Object could not be loaded as IP "
                            + templateSignature_.getInstrumentName()
                            + " v"
                            + templateSignature_.getVersion()
                            + " was not found.");
                }
                setTemplateSignature(null);
            }
        }

    }

    /**
     * Generic hook for "visitor" Design Pattern.
     * See the BusinessVisitor interface for more details.
     * 
     * @param visitor   the class whose visit() method we wish to invoke.
     * 
     * @see BusinessVisitor
     */
    public Object accept(BusinessVisitor visitor, Object arg) {
        return visitor.visit(this, arg);
    }

    /**
     * Returns the findingChart.
     * @return FindingChart[]
     */
    public FindingChart[] getFindingCharts() {
        if (findingCharts == null) {
            return new FindingChart[0];
        }
        return findingCharts;
    }

    /**
     * Sets the findingChart.
     * @param findingChart The findingChart to set
     */
    public void setFindingCharts(FindingChart[] findingCharts) {
        this.findingCharts = findingCharts;
        ObservingRun obsRun = getObsRun();
        firePropertyChangeEvent(findingChartEvent);
    }

    /**
     * Method hasFindingChartName.
     * @param string
     * @return boolean
     */
    public boolean hasFindingChartName(String filename) {
        FindingChart[] fcCharts = getFindingCharts();
        for (int i = 0; i < fcCharts.length; i++) {
            stdlog_.debug("FC Name " + fcCharts[i].getFcName());
            if (fcCharts[i].getFcName().equals(filename)) {
                return true;
            }
        }
        return false;

    }

    /**
        * Method hasFindingChartName.
        * @param string
        * @return boolean
        */
    public boolean hasFindingChartChecksum(long checksum) {
        FindingChart[] fcCharts = getFindingCharts();
        for (int i = 0; i < fcCharts.length; i++) {
            if (fcCharts[i].getCheckSum() == checksum) {
                stdlog_.debug("!!!!!!!!!!!!!!! Returning true");
                return true;
            }
        }
        return false;

    }

    /**
         * @see org.eso.ohs.dfs.StorableObject#postProecess()
         */
    public void postProecess() {
        FindingChart[] fcs = getFindingCharts();
        for (int i = 0; i < fcs.length; i++) {
            fcs[i].setObsRun(getObsRun());
        }
    }

    /**
     * Method replaceFindingChartsByName.
     * @param fcChart
     */
    public void replaceFindingChartsByName(FindingChart fcChart) {
        FindingChart[] fcCharts = getFindingCharts();
        for (int i = 0; i < fcCharts.length; i++) {
            if (fcCharts[i].getFcName() == fcChart.getFcName()) {
                fcCharts[i] = fcChart;
            }
        }
    }

    /**
     * Returns the stIntervals_.
     * @return STTimeInterval[]
     */
    public TimeInterval[] getSTTimeIntervals() {
        if(stIntervals_ == null)
            return new TimeInterval[0];
        return stIntervals_;
    }

    /**
     * Sets the stIntervals.
     * @param stIntervals_ The stIntervals_ to set
     */
    public void setSTTimeIntervals(TimeInterval[] stIntervals) {
        stIntervals_ = stIntervals;
        firePropertyChangeEvent();
    }

    /**
     * Reeturn the eph file for an OB
     * @return Ephemeris file for an ob. This is for Toos.
     */
    public EphemerisFile getEphemerisFile() {
        return ephemerisFile;
    }

    /**
     * @param file
     */
    public void setEphemerisFile(EphemerisFile file) {
        ephemerisFile = file;
        firePropertyChangeEvent();
    }
    
    public boolean hasEphemerisFile() {
        if(ephemerisFile != null && ephemerisFile.hasEphemerisFile()) {
            return true;
        }
        return false;
    }
    
    
    public boolean equals (ObservationBlock other) throws ReflectionException {
        
        // compare CalibrationBlock inherited fields
        if ( ! super.equals(other) )
            return(false);	
        
        // compare simpler fields, start with calibrationRequirements
        if ( getCalibrationRequirements().compareTo( other.getCalibrationRequirements()  )!= 0 )
            return(false);
        
        // compare target
        if ( !getTarget().equals(other.getTarget()))
            return(false);
        
        // compare acquisition template signature
        if ( !getTemplateSignature().equals(other.getTemplateSignature()) )
            return(false);
        
        // compare ephemeris file
        if (  !getEphemerisFile().equals(other.getEphemerisFile()))
            return(false);
        
        //  compare constraint sets
        
        //obtain properties from the CS to be considered in comparison
        Instrument instObj = InstrumentList.getInstance().getInstrument(this.getOd().getInstrument(),this.getOd().getIPVersion() );
        InstrumentConstraintConfig [] insConfList = instObj.getCsConfList();

		ConstraintSet cs = getConstraintSet();
		
		String [] props = new String[insConfList.length];
		
		for(int i=0;i<insConfList.length;i++) {
            String propName = insConfList[i].getName();            
            props[i] = propName;            
        }                       
        
        if ( !getConstraintSet().equals(other.getConstraintSet(), props))
            return(false);
        
           
        // iterate time intervals, order matters
        for (int i = 0; i < timeIntervals_.length; i++) {
            TimeInterval ti_this = timeIntervals_[i];
            TimeInterval ti_other = other.timeIntervals_[i];
          
            if (!ti_this.equals(ti_other))
                return(false);            
        }
        
        // iterate siderate time intervals, order also matters
        for (int i = 0; i < stIntervals_.length; i++) {
            TimeInterval ti_this = stIntervals_[i];
            TimeInterval ti_other = other.stIntervals_[i];
          
            if (!ti_this.equals(ti_other))
                return(false);            
        }  
        
        //  finding charts, order doesn't matter
        
        // if lenghts dont't differ, analyse deeper
        if (findingCharts.length == other.getFindingCharts().length) {
        
            
            
            // for each finding chart in the first OB, check for its existence in the second            
            for (int i = 0; i < findingCharts.length; i++) {
                boolean found = false;
                FindingChart fc = findingCharts[i];
                
                // find it in the 2nd OB
                for (int j = 0; j < other.getFindingCharts().length; j++) {
                    FindingChart fcOther = other.getFindingCharts()[j];
                
                    if ( fc.equals(fcOther) )
                        found = true;
                }
                
                if (!found)
                    return(false);
            }
         // if lenghts differ, return false
        }	else {
            return(false);
        }
        
                                                             
        return(true);
    }


}
