package org.eso.ohs.persistence;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.JobEventGenerator;
import org.eso.ohs.core.utilities.JobListener;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.StorableObject;
import org.eso.ohs.persistence.ObjectManager.RegistryEntry;

/**
 * Local Business Objects are stored using the AutoSaver thread.
 * 
 * <p>
 * $Id: AutoSaver.java,v 1.38 2005/02/01 14:48:02 tcanavan Exp $<br>
 * 
 * @author Bob Kemp
 */

public class AutoSaver extends Thread implements PropertyChangeListener,
        ObjectListener {

    /**
     * Standard log channel <br>
     * ChangeLog: cmavros 21-12-01 added
     */
    private static Logger stdlog_ = Logger.getLogger(AutoSaver.class);

    /** Length of a nap between polling the stayAlive_ flag */
    public static final int NAPTIME_IN_MILLISECS = 1000;

    /** Number of naps between saving queued objects to disk */
    public static final int NAPS_BETWEEN_WORK = 20;

    /** Number of naps between saving queued objects to disk */
    public static final int DELAY_AFTER_ERROR_DIALOGUE = 10000;

    /**
     * A queue of Storable Objects that will be copied to the writing queue
     * which will be periodically written to the appropriate storage media.
     */
    private Vector waitingQueue_ = new Vector();

    /**
     * A queue of Storable Objects which will be written to the selected storage
     * media
     */
    private Vector writingQueue_ = new Vector();

    /** the state of the thread: still alive or not */
    protected boolean stayAlive_ = true;

    /** Singleton instance of the AutoSaver class */
    protected static AutoSaver singleton_;

    /** Object used for locking of the writingQueue_ object */
    protected final Object WRITEQ_LOCK = new Object();

    /** Object used for locking of the waitingQueue_ object */
    protected final Object WAITQ_LOCK = new Object();

    /**
     * We need to send messages back to the GUI in case the Autosaver causes an
     * error.
     */

    private JobEventGenerator jobEventGenerator = new JobEventGenerator();

    /**
     * Private constructor: use getAutoSaver() instead.
     */
    public AutoSaver() {
        super("AutoSaver Thread");

        initialise();
    }

    protected void initialise() {
        setPriority(MIN_PRIORITY);
        start();
    }

    /**
     * Factory method to return the current instance of the AutoSaver.
     * 
     * @return the active AutoSaver instance.
     */
    public static AutoSaver getAutoSaver() {
        return singleton_;
    }

    /**
     * Utility method to set the current AutoSaver thread.
     */
    public static void setAutoSaver(AutoSaver saver) {
        singleton_ = saver;
    }

    /**
     * Start a new AutoSaver
     */
    public static void startAutoSaver() {
        ObjectManager objMgr = ObjectManager.getObjectManager();

        AutoSaver.setAutoSaver(new AutoSaver());
        objMgr.addObjectListener(AutoSaver.getAutoSaver());
    }

    /**
     * Kill any currently running AutoSaver
     */
    public static void stopAutoSaver() {
        AutoSaver saver = AutoSaver.getAutoSaver();
        ObjectManager objMgr = ObjectManager.getObjectManager();

        if (saver != null) {
            objMgr.removeObjectListener(saver);
            saver.shutdown();
            AutoSaver.setAutoSaver(null);
        }
    }

    /**
     * Save all of the queued data to disk. The thread is put into a wait state
     * while all of the data is being saved to the selected media
     * 
     * <p>
     * NOTE: all I/O should be done in the AutoSaver thread or else we must
     * cater for multiple simultaneous writes to the same file!
     */
    public void shutdown() {
        stdlog_.debug("SaveAll: telling thread to quit.");
        synchronized (WAITQ_LOCK) {
            stayAlive_ = false;

            try {
                stdlog_.debug("SaveAll: waiting...");
                WAITQ_LOCK.wait(); // wait for AutoSaver thread to finish.
                stdlog_.debug("SaveAll: all done.");
            } catch (InterruptedException e) {
                stdlog_.error(e);
            }
        }

        // At this point, the Thread is dead so no synch is needed.
        // Any object queued for saving after this point will not
        // actually be saved -- make queueing generate an error!
        waitingQueue_ = null;
        writingQueue_ = null;
        singleton_ = null;
    }

    public static boolean isActive() {
        AutoSaver saver = AutoSaver.getAutoSaver();

        if (saver == null || !saver.isAlive()) {
            return false;
        }

        synchronized (saver.WAITQ_LOCK) {
            return saver.stayAlive_;
        }
    }

    /**
     * Captures the property change event if there is any changes to storable
     * objects and inserts the Storable objects into the waiting queue.
     */
    public void propertyChange(PropertyChangeEvent ev) {
        StorableObject obj = (StorableObject) ev.getSource();
        ObjectManager objMgr = ObjectManager.getObjectManager();
        RegistryEntry regEntry = (RegistryEntry) objMgr.lookup(obj.getId());
        regEntry.setLocked(true);
        addToQueue(obj);
        //addToQueue((StorableObject) ev.getSource());
    }

    /**
     * Overridable method to determine whether an object should be queued for
     * writing to external media when it fires a PropertyChangeEvent.
     * 
     * @param ev
     *            ObjectEvent specifying the change to the Registry
     */
    protected boolean shouldAutoSave(ObjectEvent ev) {
        Media device = ev.getMedia();
        boolean result = device.equals(Media.LOCAL);

        if (Config.getCfg().isInOperationsMode()) {
            return false;
        }

        return result;
    }

    /**
     * Adds a PropertyChangeListener event to the StorableObject which is
     * contained within is initialised from the ObjectEvent
     */
    public void objectAdded(ObjectEvent ev) {
        if (shouldAutoSave(ev)) {
            StorableObject sobj = ev.getStorableObject();
            sobj.addPropertyChangeListener(this);
            if (ev.getType() == ObjectEvent.OBJECT_CREATED) {
                ObjectManager objMgr = ObjectManager.getObjectManager();
                RegistryEntry regEntry = (RegistryEntry) objMgr.lookup(sobj
                        .getId());
                regEntry.setLocked(true);
                addToQueue(sobj); // write new objects immediately
            }
        }
    }

    /**
     * An object has been renamed but we don't care. It gets handled in
     * ObjectManager.
     */
    public void objectMoved(ObjectEvent ev) {
    }

    /**
     * Remove the propertyChangeListener event from the Storable Object and from
     * the queue.
     */
    public void objectRemoved(ObjectEvent ev) {
        StorableObject sobj = ev.getStorableObject();

        if (sobj != null) {
            sobj.removePropertyChangeListener(this);
            removeFromQueue(sobj);
        }
    }

    public void objectRemovedFromRegistry(ObjectEvent ev) {

    }

    /**
     * Add a storable object to the waiting queue
     * 
     * @param bo
     *            StorableObject to queue for saving
     */
    public void addToQueue(StorableObject bo) {
        if (bo == null)
            stdlog_
                    .debug("Object is Garbage collected cannot save this object");
        stdlog_.debug("Queueing: " + bo);

        synchronized (WAITQ_LOCK) {
            if (stayAlive_ && !waitingQueue_.contains(bo)) {
                waitingQueue_.addElement(bo);
            }
        }
    }

    /**
     * Remove a StorableObject from the queues to be saved.
     * 
     * @param bo
     *            StorableObject to dequeue
     */
    public void removeFromQueue(StorableObject bo) {
        //stdlog_.debug("Dequeueing: " + bo);

        synchronized (WAITQ_LOCK) {
            synchronized (WRITEQ_LOCK) {
                waitingQueue_.removeElement(bo);

                int size = writingQueue_.size();

                for (int j = 0; j < size; j++) {
                    if (bo == writingQueue_.elementAt(j)) {
                        stdlog_.debug("Got it: " + bo);
                        writingQueue_.setElementAt(null, j);
                    }
                }
            }
        }

    }

    private void shortPause() {
        try {
            for (int napCount = 1; napCount <= NAPS_BETWEEN_WORK && stayAlive_; napCount++) {
                sleep(NAPTIME_IN_MILLISECS);
            }
        } catch (InterruptedException e) {
            // don't care
        }
    }

    /**
     * This is the main method of the thread. Waiting queue gets copied to the
     * writingQueue_ and the writingQueue_ gets stored to the appropriate Media.
     * When the writingQueue_ has finished been written all of its elements will
     * be removed.
     * 
     * <b>Must be made utterly bulletproof: if the thread dies unexpectedly,
     * nothing gets saved! </b>
     */
    public void run() {
        Vector queue;

        for (;;) {
            shortPause();

            synchronized (WAITQ_LOCK) {
                synchronized (WRITEQ_LOCK) {

                    if (!stayAlive_)
                        break;

                    // Swap queues around,
                    queue = waitingQueue_;
                    waitingQueue_ = writingQueue_;
                    writingQueue_ = queue;
                }
            }

            storeQueue(writingQueue_);
            synchronized (WRITEQ_LOCK) {
                writingQueue_.removeAllElements();
            }
        }

        storeQueue(waitingQueue_);

        stdlog_.debug("Goodbye Cruel World!");
        synchronized (WAITQ_LOCK) {
            // notify (all) processes that called saveAll()
            // of our imminent demise.
            WAITQ_LOCK.notifyAll();
        }
        stdlog_.debug("Splat!");
    }

    /**
     * Store the queue of Storable objects to the appropriate Media. Gets a
     * reference to the singleton object manager class and call putBusObj which
     * registers the StorableObject with the ObjectManager and writes it to the
     * selected media
     * 
     * @param queue
     *            Queue of objects to be writen to external media
     */
    private void storeQueue(Vector queue) {
        int size = queue.size();
        StorableObject sobj = null;
        ObjectManager objMgr = ObjectManager.getObjectManager();

        for (int j = 0; j < size; j++) {
            try {
                synchronized (WRITEQ_LOCK) {
                    sobj = (StorableObject) queue.elementAt(j);
                    stdlog_.debug("Saving: " + sobj);
                    if (sobj != null) {
                        objMgr.putBusObj(sobj);
                    }

                }
            } catch (ObjectIOException e) {
                stdlog_.error(e);
                jobEventGenerator.fireOnException(e);
                //Config.getCfg().alertUser(e);
                // XXX requeue it
                // Needs to do something better!
                // For example, _must_ notify user
                addToQueue(sobj);
            } catch (IllegalArgumentException e) {
                stdlog_.error("Not Saving: " + sobj.getId());
            } catch (Exception e) {
                stdlog_.error("UNEXPECTED EXCEPTION while saving "
                        + sobj.getId());
                // XXX should handle this error better!
                stdlog_.error(e);
                jobEventGenerator.fireOnException(e);
                //Config.getCfg().alertUser(e);
            } catch (OutOfMemoryError e) {
                stdlog_.error(e);
                jobEventGenerator.fireOnOutOfMemory(e);
                if (sobj != null) {
                    //addToQueue(sobj);
                }
                shortPause();
            }
        }
    }


    /**
     * Returns the jobEventGenerator.
     * @return JobEventGenerator
     */
    public void addJobEvent(JobListener listener) {
        jobEventGenerator.addJobListener(listener);
    }

    /**
     * Sets the jobEventGenerator.
     * @param jobEventGenerator The jobEventGenerator to set
     */
    public void removeJobEvent(JobListener listener) {
        jobEventGenerator.removeJobListener(listener);

    }
    
}