package org.eso.ohs.persistence.dbase.phase2;

import java.io.File;
import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.ConnectionPool;
import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.instruments.Instrument;
import org.eso.ohs.instruments.InstrumentException;
import org.eso.ohs.instruments.InstrumentList;
import org.eso.ohs.persistence.Config;


/** 
 * This class instanciates a Singleton Reference to a List of
 * Instrument objects. The class extends InstrumentList, and 
 * adds the ability to download IPs from the database if the
 * IP is not found locally.
 * 
 * @version $Revision: 1.1 $
 * @author Nick Kornweibel for ESO. 
 */

public class DBInstrumentList extends InstrumentList {

    private String ipCacheDir_ = null;
    private String instDir_ = null;
    private boolean ipMode_;

    private String dbUserName_ = null;
    private String dbPassword_ = null;
    
    /**
     * Standard log channel 
     */
    private static Logger stdlog_ = Logger.getLogger( DBInstrumentList.class );
    
    protected DBInstrumentList(boolean ipMode,String ipc,String instDir, String dbUserName, String dbPassword) {
        super();
        ipCacheDir_ = ipc;
        ipMode_ = ipMode;
        instDir_ = instDir;

        dbUserName_ = dbUserName;
        dbPassword_ = dbPassword;
        if (ipCacheDir_ != null && ipCacheDir_.length() > 0) {
            File f = new File(ipCacheDir_);
            ipCacheDir_ = f.getAbsolutePath();
        } else {
            ipCacheDir_ = null;
        }
    }

    /**
     * @return the instrument requested, with the version requested.
     * If the instrument is not found in the local cache, the database
     * will be queried for the IP, which will be downloaded and extracted
     * if located.
     *
     * NOTE:
     * Unfortunately it is unclear what to do in the case of error...if an
     * instrument is required from the database, but fails to be downloaded,
     * then it is handled as if the IP simply did not exist and null is returned.
     */
    public Instrument getInstrument(String inst, float version) {
        Instrument rc = super.getInstrument(inst, version);
        if (rc == null) {
            stdlog_.debug("Download in DBInstrumentList.getInstrument().");
            // query the DB and download the IP if it exists.
            float v = version;
            if (version < 0) {
                v = 0;
            }
            if (downloadInstrumentPackage(inst, v)) {
                try {
                    InstrumentList.reloadInstrumentList(ipMode_,instDir_,ipCacheDir_);
                } catch (InstrumentException e) {
                    rc = null;
                }
                rc = InstrumentList.getInstance().getInstrument(inst, version);
            }
        }
        stdlog_.debug("Exit DBInstrumentList.getInstrument().");
        return rc;
    }
    
    /**
     * Download and extract the specified IP from the database into the IPCache directory,
     * and add it to the InstrumentList.
     * There is no check if the IP is aleary in the cache.
     *
     * NOTE:
     * Unfortunately it is unclear what to do in the case of error...so no exception is
     * thrown - a boolean is returned to indicate that if the IP was downloaded or not,
     * not distinguishing between an IP not in the database or one that failed to be
     * extracted.
     *
     * A seperate connection object is opened and closed for this operation. It is
     * outside of the connection pool and is used simple to download the IP. It is out
     * side of the connection pool as the connection to the DB for downloading IPs
     * is compeletely hidden from the caller.
     */
    private boolean downloadInstrumentPackage(String instrument, float version) {
        boolean rc = false;
        if (ipCacheDir_ != null) {
            Connection con = null;
            try {
                Config cfg = Config.getCfg();
                con = ConnectionPool.getPool().openExtraPoolConnection(
                    cfg.getUrl(ObservationBlock.getSuffix()), dbUserName_, dbPassword_);
                
                InstrumentPackageDBIO.getIP(instrument, version, ipCacheDir_, true, con);
                con.close();
                con = null;
                rc = true;
            } catch (IOException e1) {
                stdlog_.error("Failed downloading IP: "+e1.toString(), e1);
                rc = false;
            } catch (SQLException e2) {
                stdlog_.error("SQL Error downloading IP: "+e2.toString(), e2);
                rc = false;
            } finally {
                if (con != null) {
                    try {
                        con.close();
                    } catch (SQLException e) {
                        stdlog_.error("Failed closing IP DB connection: "+e.toString(), e);
                    }
                }
            }
            
        }
        return rc;
    }
}
    