package org.eso.ohs.phase2.apps.p2pp.actions;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Toolkit;

import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.SwingUtilities;

import org.eso.ohs.core.gui.widgets.ErrorMessages;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.core.utilities.SystemStatusEvent;
import org.eso.ohs.persistence.Config;
import org.eso.ohs.persistence.Media;
import org.eso.ohs.persistence.ObjectNotFoundException;
import org.eso.ohs.phase2.apps.p2pp.DbaseWindow;
import org.eso.ohs.phase2.apps.p2pp.MainWindow;
import org.eso.ohs.phase2.apps.p2pp.MenuGroups;
import org.eso.ohs.phase2.apps.p2pp.P2PP;
import org.eso.ohs.phase2.apps.p2pp.Persistence;

/**
 * This is a P2PP specific utility class to restart the P2PP
 * windows.
 * It is used after major data manipulating operations such as
 * downloading obsruns, changing user, etc.
 *
 * This class was a private class of the ReloginAction, but has
 * been made a seperate class so that it may be used elsewhere.
 *
 * To use.include it in code similar to the following:
 *
 *      try {
 *          // when changing user id....
 *          SwingUtilities.invokeLater(new RestartWindows(oldUserId, status));
 *          // otherwise....
 *          SwingUtilities.invokeLater(new RestartWindows());
 *      } catch (ObjectIOException e) {
 *          report error message.
 *      }
 * <br>
 * @version $Revision: $
 */
public class RestartWindows implements Runnable {

    public static final String cvsID_ = "$Id: $";

    private JFrame inProgressMsg_;

    private int status_;

    private int oldUserId_ = -1;

    private int progressPoint_ = 0;


    /**
     * Use this constructor in the case where the user id changes
     * (and the local storage must be changed.
     */
    public RestartWindows(int oldUserId, int status) {
        oldUserId_ = oldUserId;
        status_ = status;
    }

    /**
     * Standard constructor - sued when windows restart
     * is required with no user id change.
     */
    public RestartWindows() {
        // nothing to do.
    }

    public void run() {
        Persistence objMgr = Persistence.getInstance();

        /*
         * NOTE:
         * The restart is implemented as a finite-state automaton
         * to give the Java runtime system multiple opportunities
         * to redraw the InProgressMsg_ frame.  Otherwise, it keeps
         * getting hidden behind other windows.
         */
        //System.out.println("CLASS ReloginAction(Restart) - run");
        try {
            switch (progressPoint_) {
                case 0 :
                    inProgressMsg_ = createInProgressWindow();
                    inProgressMsg_.setVisible(true);
                    MainWindow.getView().hideViews();
                    break;

                case 1 :
                    // only do this if the user id has changed.
                    P2PP.fireSystemStatusEvent(SystemStatusEvent.SYNC, null);
                    P2PP.lockMgr.unlockLocalStorage(oldUserId_);

                    objMgr.restart();
                    objMgr = Persistence.getInstance();
                    objMgr.setUserId(Config.getCfg().getUserName());
                    break;

                case 2 :
                    // only do this if the user id has changes and
                    // we are not in ops mode.
                    if (oldUserId_ != -1
                        && status_ == 0
                        && !Config.getCfg().isInOperationsMode()) {
                        P2PP.downloadObservingRuns();
                    }
                    break;

                case 3 :
                    P2PP.fireSystemStatusEvent(SystemStatusEvent.RELOGIN, null);
                    break;

                case 4 :
                    restartMainWindows();
                    disposeOfProgressWindow();
                    return;
            }

        } catch (ObjectIOException e) {
            disposeOfProgressWindow();
            ErrorMessages.announceIOError(MainWindow.getPrincipalWindow(), e);
            return; // XXX Should quit application???

        } catch (ObjectNotFoundException e) {
            disposeOfProgressWindow();
            ErrorMessages.announceNoObject(MainWindow.getPrincipalWindow(), e);
            return; // XXX Should quit application???
        }

        // Will bring window to the front!
        inProgressMsg_.setVisible(true);
        progressPoint_++;
        SwingUtilities.invokeLater(this);
    }

    private void disposeOfProgressWindow() {
        inProgressMsg_.setVisible(false);
        inProgressMsg_.dispose();
    }

    public JFrame createInProgressWindow() {
        JFrame frm = new JFrame("Login in progress...");
        Container cont = frm.getContentPane();
        JLabel lab = new JLabel("In Progress... Please wait");

        cont.setLayout(new BorderLayout());
        cont.add(lab, BorderLayout.CENTER);

        lab.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));

        Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();

        frm.setSize(300, 100);
        frm.setLocation(
            (screen.width - frm.getSize().width) / 2,
            (screen.height - frm.getSize().height) / 2);

        return frm;
    }

    /**
     * Closes all MainWindow instances, eg because the userId
     * has changed and a completely different set of BOs must
     * be displayed.
     *
     * @exception ObjectIOException Object I/O Exception
     */
    public void restartMainWindows() throws ObjectIOException {
        MenuGroups[] mgrps = MenuGroups.getWindowList();
        MainWindow wdw;

        wdw = MainWindow.getPrincipalWindow();
        if (wdw != null) {
            // Make all existing windows invisible.
            for (int j = 0; j < mgrps.length; j++) {
                //System.out.println("CLASS ReloginAction(Restart) - restartMainWindows: Window = " + mgrps[j].getClass().getName());
                mgrps[j].setDefaultCloseOperation(JFrame.HIDE_ON_CLOSE);
                mgrps[j].setVisible(false);
            }
        }
        if (Config.getCfg().isInOperationsMode()) {
            DbaseWindow dbWin = DbaseWindow.getDbaseWindow(Media.DBASE);
            P2PP.setScreenSize(dbWin);
            dbWin.setVisible(true);
        } else {
            wdw = new MainWindow(Media.LOCAL);
            wdw.setVisible(true);
        }

        // Make all existing windows invisible.
        for (int j = 0; j < mgrps.length; j++) {
            mgrps[j].dispose();
        }
    }
}

